unit Chinese;

{ An other game component by Paul TOTH

Copyright (C) 1999 Paul TOTH <tothpaul@multimania.com>
 http://www.multimania.com/tothpaul

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
}

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs;

type
 TCell=(clBorder,clEmpty,clPawn,clMove,clEat,clJump);
 TBoard=array[-1..+9,-1..+9] of TCell;
      {     1234567   }
      {   XXXXXXXXXXX }
      {   XXXXXXXXXXX }
      { 1 XXXXOOOXXXX }
      { 2 XXXXOOOXXXX }
      { 3 XXOOOOOOOXX }
      { 4 XXOOO OOOXX }
      { 5 XXOOOOOOOXX }
      { 6 XXXXOOOXXXX }
      { 7 XXXXOOOXXXX }
      {   XXXXXXXXXXX }
      {   XXXXXXXXXXX }

 TMove=record
  Start,Target:record x,y:integer end;
 end;

 TChinese = class(TCustomControl)
  private
   fRgn:array[0..11] of TPoint;
   fMoves:array[1..32] of TMove;
   fMoveCount:integer;
   fBoard:TBoard;
   fPawn:TColor;
   fHole:TColor;
   fPawns:integer;
   fCellWidth:integer;
   fCellHeight:integer;
   fCellLeft:integer;
   fCellTop:integer;
   fCellX:integer;
   fCellY:integer;
   fSelect:boolean;
   fSelectX:integer;
   fSelectY:integer;
   procedure WMSize(var Message: TWMSize); message WM_SIZE;
  protected
   constructor Create(AOwner:TComponent); override;
   procedure Paint; override;
   procedure MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: Integer); override;
   procedure MouseMove(Shift: TShiftState; X, Y: Integer); override;
   procedure MouseUp(Button: TMouseButton; Shift: TShiftState; X, Y: Integer); override;
   function ThisWay(Way:integer):boolean;
   function CanPlayXY:integer;
   procedure SetCell(x,y:integer; CellState:TCell);
   procedure Play;
   procedure Select;
   procedure UnSelect;
   procedure DoMove(x1,y1,x2,y2:integer);
  public
   procedure NewGame;
   procedure Undo;
  published
   property Pawn:TColor read fPawn write fPawn default clRed;
   property Hole:TColor read fHole write fHole default clBlack;
   property Color default clMaroon;
   property Align;
  end;

procedure Register;

implementation

procedure Register;
begin
  RegisterComponents('MySoft', [TChinese]);
end;

Constructor TChinese.Create(AOwner:TComponent);
 begin
  inherited Create(AOwner);
  fPawn:=clRed;
  fHole:=clBlack;
  Color:=clMaroon;
  NewGame;
 end;

procedure TChinese.WMSize(var Message: TWMSize);
 var
  x:integer;
 begin
  inherited;
  fCellWidth :=ClientWidth div 7;
  fCellHeight:=ClientHeight div 7;
  if (fCellWidth=0)or(fCellHeight=0) then exit;
  fCellLeft:=(ClientWidth-7*fCellWidth) div 2;
  fCellTop:=(ClientHeight-7*fCellHeight) div 2;
  fRgn[0].x:=fCellLeft+2*fCellWidth;
  fRgn[0].y:=fCellTop;
  fRgn[1].x:=fCellLeft+5*fCellWidth;
  fRgn[1].y:=fCellTop;
  fRgn[2].x:=fCellLeft+5*fCellWidth;
  fRgn[2].y:=fCellTop+2*fCellHeight;
  fRgn[3].x:=fCellLeft+7*fCellWidth;
  fRgn[3].y:=fCellTop+2*fCellHeight;
  fRgn[4].x:=fCellLeft+7*fCellWidth;
  fRgn[4].y:=fCellTop+5*fCellHeight;
  fRgn[5].x:=fCellLeft+5*fCellWidth;
  fRgn[5].y:=fCellTop+5*fCellHeight;
  fRgn[6].x:=fCellLeft+5*fCellWidth;
  fRgn[6].y:=fCellTop+7*fCellHeight;
  fRgn[7].x:=fCellLeft+2*fCellWidth;
  fRgn[7].y:=fCellTop+7*fCellHeight;
  fRgn[8].x:=fCellLeft+2*fCellWidth;
  fRgn[8].y:=fCellTop+5*fCellHeight;
  fRgn[9].x:=fCellLeft;
  fRgn[9].y:=fCellTop+5*fCellHeight;
  fRgn[10].x:=fCellLeft;
  fRgn[10].y:=fCellTop+2*fCellHeight;
  fRgn[11].x:=fCellLeft+2*fCellWidth;
  fRgn[11].y:=fCellTop+2*fCellHeight;
  SetWindowRgn(Handle,CreatePolygonRgn(fRgn,12,ALTERNATE),True);
  for x:=1 to 6 do dec(fRgn[x].x);
  for x:=4 to 9 do dec(fRgn[x].y);
 end;

procedure TChinese.Paint;
 var
  x,y:integer;
  z:TCell;
  ox,oy:integer;
  dx,dy:integer;
 begin
  if (fCellWidth=0)or(fCellHeight=0) then exit;
  With Canvas do begin
   Brush.Color:=Self.Color;
   Polygon(fRgn);
   oy:=fCellTop;
   dx:=fCellWidth div 8;
   dy:=fCellHeight div 8;
   for y:=1 to 7 do begin
    ox:=fCellLeft;
    for x:=1 to 7 do begin
     z:=fBoard[x,y];
     if z>clBorder then begin
      case z of
       clEmpty:Brush.Color:=fHole;
       clPawn:Brush.Color:=fPawn;
       else Brush.Color:=self.color;
      end;
      Ellipse(ox+dx,oy+dy,ox+fCellWidth-dx,oy+fCellHeight-dy);
      if z=clMove then begin
       Brush.Color:=fPawn;
       Ellipse(ox+2*dx,oy+2*dy,ox+fCellWidth-2*dx,oy+fCellHeight-2*dy);
      end;
     end;
     inc(ox,fCellWidth);
    end;
    inc(oy,fCellHeight);
   end;
  end;
 end;

procedure TChinese.MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
 begin
  if (fCellWidth=0)or(fCellHeight=0) then exit;
  SetCaptureControl(Self);
 end;

procedure TChinese.MouseMove(Shift: TShiftState; X, Y: Integer);
 begin
  if (fCellWidth>0)and(fCellHeight>0) then begin
   fCellX:=1+(X-fCellLeft) div fCellWidth;
   fCellY:=1+(Y-fCellTop) div fCellHeight;
   if CanPlayXY>0 then Cursor:=crCross else Cursor:=crDefault;
  end;
 end;

procedure TChinese.MouseUp(Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
 begin
  SetCaptureControl(nil);
  if Button=mbRight then UnSelect else  begin
   fCellX:=1+(X-fCellLeft) div fCellWidth;
   fCellY:=1+(Y-fCellTop) div fCellHeight;
   case CanPlayXY of
    0:if fSelect then UnSelect;
    1:Play;
    else Select;
   end;
  end;
 end;

Const
 Mvt:array[0..3,0..1] of integer=(
          ( 0,-1),
  (-1, 0),        (+1, 0),
          ( 0,+1)
 );

function TChinese.ThisWay(Way:integer):boolean;
 var
  dx,dy:integer;
 begin
  dx:=Mvt[Way,0];
  dy:=Mvt[Way,1];
  Result:=(fBoard[fCellX+dx,fCellY+dy]=clPawn)
       and(fBoard[fCellX+2*dx,fCellY+2*dy]=clEmpty);
 end;

function TChinese.CanPlayXY:integer;
 var
  m:integer;
 begin
  Result:=0;
  if fSelect then begin
   if (fBoard[fCellX,fCellY]=clEmpty) then begin
    if ((fCellX=fSelectX) and (Abs(fCellY-fSelectY)=2))
    or ((fCellY=fSelectY) and (Abs(fCellX-fSelectX)=2)) then Inc(Result);
   end;
  end else begin
   if (fBoard[fCellX,fCellY]=clPawn) then begin
    for m:=0 to 3 do if ThisWay(m) then inc(Result);
   end;
  end;
 end;

procedure TChinese.SetCell(x,y:integer;CellState:TCell);
 Var
  R:TRect;
 begin
  fBoard[x,y]:=CellState;
  R.Right:=fCellLeft+x*fCellWidth;
  R.Bottom:=fCellTop+y*fCellHeight;
  R.Left:=R.Right-fCellWidth;
  R.Top:=R.Bottom-fCellHeight;
  InvalidateRect(Handle,@R,True);
 end;

procedure TChinese.Play;
 var
  m,dx,dy:integer;
 begin
  if fSelect then begin
   DoMove(fSelectX,fSelectY,fCellX,fCellY);
   fSelect:=False;
  end else begin
   m:=0; While Not ThisWay(m) do inc(m);
   dx:=2*Mvt[m,0];
   dy:=2*Mvt[m,1];
   DoMove(fCellX,fCellY,fCellX+dx,fCellY+dy);
  end;
  if fPawns=0 then begin
   if fBoard[4,4]=clPawn then ShowMessage('You win !');
  end;
 end;

procedure TChinese.Select;
 begin
  fSelectX:=fCellX;
  fSelectY:=fCellY;
  SetCell(fSelectX,fSelectY,clMove);
  fSelect:=True;
 end;

procedure TChinese.UnSelect;
 begin
  if fSelect then begin
   SetCell(fSelectX,fSelectY,clPawn);
   fSelect:=False;
  end;
 end;

Procedure TChinese.NewGame;
 var
  x:integer;
 begin
  FillChar(fBoard,sizeof(fBoard),byte(clBorder));
  For x:=1 to 2 do FillChar(fBoard[x,3],3,byte(clPawn));
  For x:=3 to 5 do FillChar(fBoard[x,1],7,byte(clPawn));
  For x:=6 to 7 do FillChar(fBoard[x,3],3,byte(clPawn));
  fBoard[4,4]:=clEmpty;
  fPawns:=32;
  fSelect:=False;
  fMoveCount:=0;
  Invalidate;
 end;

procedure TChinese.DoMove(x1,y1,x2,y2:integer);
 begin
  inc(fMoveCount);
  SetCell(x1,y1,clEmpty);
  SetCell(x2,y2,clPawn);
  with fMoves[fMoveCount] do begin
   Start.x:=x1;
   Start.y:=y1;
   Target.x:=x2;
   Target.y:=y2;
  end;
  if x2>x1 then inc(x1) else if x2<x1 then dec(x1);
  if y2>y1 then inc(y1) else if y2<y1 then dec(y1);
  SetCell(x1,y1,clEmpty);
  dec(fPawns);
 end;

procedure TChinese.Undo;
 var
  x1,y1,x2,y2:integer;
 begin
  if fMoveCount=0 then exit;
  with fMoves[fMoveCount] do begin
   x1:=Start.x;
   y1:=Start.y;
   x2:=Target.x;
   y2:=Target.y;
  end;
  SetCell(x1,y1,clPawn);
  SetCell(x2,y2,clEmpty);
  if x2>x1 then inc(x1) else if x2<x1 then dec(x1);
  if y2>y1 then inc(y1) else if y2<y1 then dec(y1);
  SetCell(x1,y1,clPawn);
  dec(fMoveCount);
  inc(fPawns);
 end;

end.
